"""Create a tkGUI controller"""

from ..config import Mode
from .tkGUI.modes import ControllerSwept, ControllerRT
from .tkGUI.modes import args_swept, args_rt

def GetController(mode):
    if mode == Mode.SWEPT:
        return ControllerSwept
    elif mode == Mode.RT:
        return ControllerRT

def ControllerArgs(mode):
    if mode == Mode.SWEPT:
        return args_swept
    elif mode == Mode.RT:
        return args_rt



import argparse
import sys

from .. import err
from ..specan import SpecAn

from ..config import config, Mode, View
from ..model.model import ModelArgs
from ..controller.controller import ControllerArgs

def define_args():
    parser = argparse.ArgumentParser("pyspecan", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    mon = parser.add_argument_group("developer toggles")
    mon.add_argument("--mon_mem", action="store_true")
    mon.add_argument("--profile", action="store_true")
    return parser

def _main(args):
    SpecAn(**vars(args))

def _process_args(parser):
    run_help = False
    if "-h" in sys.argv:
        run_help = True
        sys.argv.pop(sys.argv.index("-h"))
    elif "--help" in sys.argv:
        run_help = True
        sys.argv.pop(sys.argv.index("--help"))
    args, remaining = parser.parse_known_args()
    mode = Mode.get(args.mode)

    if mode == Mode.NONE:
        raise err.UnknownOption(f"Unknown mode {args.mode}")

    view = View.get(args.view)
    if view == View.NONE:
        raise err.UnknownOption(f"Unknown view {args.view}")

    if args.mon_mem:
        config.MON_MEM = True
    if args.profile:
        config.PROFILE = True

    ModelArgs(mode)(parser)
    ControllerArgs(view, mode)(parser)

    args = parser.parse_args()
    if run_help:
        parser.print_help()
        exit()
    return args

def main():
    parser = define_args()
    parser.add_argument("-v", "--view", type=str, default=View.tkGUI.name, choices=View.choices())
    parser.add_argument("-m", "--mode", type=str.upper, default=Mode.SWEPT.name, choices=Mode.choices())
    _main(_process_args(parser))



"""Base Views for tkGUI View plots"""
import tkinter as tk
from tkinter import ttk

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
# from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk

from ...backend.mpl.plot import _Plot, Plot, BlitPlot


class GUIPlot:
    """tkinter wrapper for pyspecan.backend.mpl.plot"""
    __slots__ = (
        "view", "_root", "plotter", "settings", "ready",
        "fr_main", "fr_canv", "fr_sets", "btn_toggle",
        "wg_sets",
    )
    def __init__(self, view, root, fig: Figure, plotter=_Plot):
        if plotter is _Plot:
            plotter = Plot
        self.view = view
        self._root = root
        self.settings = {}
        self.ready = False

        self.fr_main = ttk.Frame(root)

        self.fr_sets = ttk.Frame(self.fr_main)
        self.wg_sets = {}
        self.draw_settings(self.fr_sets)
        self.fr_sets.pack(side=tk.LEFT, fill=tk.Y)
        self.fr_sets.pack_forget()

        self.fr_canv = ttk.Frame(self.fr_main)
        self.fr_canv.pack(fill=tk.BOTH, expand=True)
        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = plotter(fig)
        # toolbar = NavigationToolbar2Tk(canvas, root)
        # toolbar.update()
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore

        self.btn_toggle = ttk.Button(self.fr_canv, text="Settings", style="Settings.TButton")
        self.btn_toggle.place(relx=0.0, rely=0.0, width=50, height=25)

        self.fr_main.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    @property
    def fig(self):
        return self.plotter.fig
    def ax(self, name):
        return self.plotter.ax(name)
    def add_ax(self, *args, **kwargs):
        return self.plotter.add_ax(*args,**kwargs)

    def draw_settings(self, parent, row=0):
        """Initialize settings panel"""
        raise NotImplementedError()

class GUIBlitPlot(GUIPlot):
    """tkinter wrapper for pyspecan.plot.mpl BlitPlot"""
    def __init__(self, view, root, fig):
        super().__init__(view, root, fig, BlitPlot)


class GUIFreqPlot(GUIBlitPlot):
    """Frequency domain view helpers"""
    __slots__ = ("lbl_lo", "lbl_hi")
    def __init__(self, view, root, fig):
        super().__init__(view, root, fig)

        self.lbl_lo = ttk.Label(self.fr_canv, text="V")
        self.lbl_hi = ttk.Label(self.fr_canv, text="^")

    def draw_settings(self, parent, row=0):
        var_scale = tk.StringVar(self.fr_sets)
        ent_scale = ttk.Entry(self.fr_sets, textvariable=var_scale, width=10)

        var_ref_level = tk.StringVar(self.fr_sets)
        ent_ref_level = ttk.Entry(self.fr_sets, textvariable=var_ref_level, width=10)

        var_vbw = tk.StringVar(self.fr_sets)
        ent_vbw = ttk.Entry(self.fr_sets, textvariable=var_vbw, width=10)

        var_window = tk.StringVar(self.fr_sets)
        cb_window = ttk.Combobox(self.fr_sets, textvariable=var_window, width=9)

        self.wg_sets["scale"] = ent_scale
        self.settings["scale"] = var_scale
        self.wg_sets["ref_level"] = ent_ref_level
        self.settings["ref_level"] = var_ref_level
        self.wg_sets["vbw"] = ent_vbw
        self.settings["vbw"] = var_vbw
        self.wg_sets["window"] = cb_window
        self.settings["window"] = var_window

        ttk.Label(parent, text="Scale/Div").grid(row=row, column=0)
        ent_scale.grid(row=row, column=1)
        row += 1
        ttk.Label(parent, text="Ref Level").grid(row=row, column=0)
        ent_ref_level.grid(row=row, column=1)
        row += 1
        ttk.Label(parent, text="VBW").grid(row=row, column=0)
        ent_vbw.grid(row=row, column=1)
        row += 1
        ttk.Label(parent, text="Window").grid(row=row, column=0)
        cb_window.grid(row=row, column=1)
        row += 1
        return row



"""Initialize tkGUI Controller"""
import math
import threading
import time
import datetime as dt

import tkinter as tk

from ..base import Controller as _Controller

from ...config import config, Mode

from ...utils import dialog
from ...utils.time import strfmt_td
from ...utils.monitor import Memory

from ...model.model import Model
from ...model.reader import Format
from ...view.tkGUI.base import View as GUI

from ...backend.mpl.plot import BlitPlot
from ...backend.tk import theme as theme_tk

from .panels import PanelController
from .plot_base import FreqPlotController, BlitPlot

def define_args(parser):
    ctrl = parser.add_argument_group("tkGUI")
    ctrl.add_argument("--theme", default="Dark", choices=[k for k in theme_tk.theme.keys()])
    ctrl.add_argument("--sweep", default=50.0, type=float)
    ctrl.add_argument("--show", default=50.0, type=float)
    return ctrl

class Controller(_Controller):
    """tkGUI Controller"""
    def __init__(self, model: Model, view: GUI, **kwargs):
        super().__init__(model, view)
        self.view: GUI = self.view # type hints
        self.running = False
        self._stop = False
        self.time_show = kwargs.get("show", 50.0)
        self.nfft_exp = int(math.log2(self.model.nfft))
        self.model.set_sweep_time(kwargs.get("sweep", 50.0))
        self._last_f = None
        self.panel: PanelController = None # type: ignore

        self.view.sld_samp.scale.config(from_=0, to=self.model.reader.max_samp) # resolution=self.model.block_size
        self.view.sld_samp.scale.config(command=self.handle_sld_samp)

        self.view.ent_sweep.bind("<Return>", self.handle_event)
        self.view.var_sweep.set(f"{self.model.sweep_time:02.3f}")
        self.view.ent_show.bind("<Return>", self.handle_event)
        self.view.var_show.set(f"{self.time_show:02.3f}")

        self.view.btn_prev.config(command=self.prev)
        self.view.btn_next.config(command=self.next)
        self.view.btn_start.config(command=self.start)
        self.view.btn_stop.config(command=self.stop)
        self.view.btn_reset.config(command=self.reset)
        self.view.var_draw_time.set(f"{0.0:06.3f}s")

        self.view.btn_file.config(command=self.handle_btn_file)
        self.view.cb_file_fmt.config(values=list([v.name for v in Format]))
        self.view.cb_file_fmt.bind("<<ComboboxSelected>>", self.handle_event)
        self.view.ent_fs.bind("<Return>", self.handle_event)
        self.view.ent_cf.bind("<Return>", self.handle_event)
        self.view.ent_nfft_exp.bind("<Return>", self.handle_event)

        self.thread: threading.Thread = None # type: ignore

        style = kwargs.get("theme", "Dark")

        theme_tk.get(style)(self.view.root) # pyright: ignore[reportCallIssue]

    def start(self):
        if self.running:
            return
        if self.model.reader.path is None:
            return
        self.running = True
        self.view.btn_start.config(state=tk.DISABLED)
        self.view.btn_stop.config(state=tk.ACTIVE)
        self.thread = threading.Thread(target=self.loop)
        self.thread.start()

    def stop(self):
        if not self._stop and not self.running:
            return
        self.running = False
        self.view.btn_stop.config(state=tk.DISABLED)
        self.view.btn_start.config(state=tk.ACTIVE)
        self.thread.join(timeout=0.2)

    def reset(self):
        self.stop()
        self.model.reset()
        self.panel.reset()
        self.draw_tb()

    def prev(self):
        self.stop()
        return self._prev()

    def next(self):
        self.stop()
        return self._next()

    def _prev(self):
        valid = self.model.prev()
        tplot = None
        if valid:
            tplot = self._plot()
        return (valid, tplot)

    def _next(self):
        valid = self.model.next()
        tplot = None
        if valid:
            tplot = self._plot()
        return (valid, tplot)

    def loop(self):
        while self.running:
            time_show = self.time_show/1000 # convert ms to s
            valid, ptime = self._next()
            if not valid or ptime is None:
                break
            wait = time_show-ptime
            if wait > 0:
                self.view.lbl_msg.configure(text="")
                time.sleep(wait)
            else:
                self.model.skip_time(-wait)
                self.view.lbl_msg.configure(text="OVERFLOW")

    def _plot(self):
        if config.MON_MEM:
            Memory().peak()
        ptime = time.perf_counter()
        self._check_f()
        self.on_plot()
        ptime = (time.perf_counter() - ptime)
        self.view.var_draw_time.set(f"{ptime:06.3f}s")
        self.draw_tb()
        # print(f"Plotted in {ptime*1000:.1f}ms / {self.time_show}")
        return ptime

    def on_plot(self):
        for child in self.panel.rows:
            for pane in self.panel.cols[child]:
                view = self.panel.view[child][pane]
                if view is None:
                    continue
                if not isinstance(view.plotter, BlitPlot):
                    view.plotter.cla()
                    print("Cleared plot!")
                if isinstance(view, FreqPlotController):
                    vbw = float(pane.sets["vbw"].get())
                    window = pane.sets["window"].get()
                    view.plot(self.model.f, self.model.psd(vbw, window))

    def on_update_f(self, f):
        for child in self.panel.rows:
            for pane in self.panel.cols[child]:
                view = self.panel.view[child][pane]
                if view is None:
                    continue
                if isinstance(view, FreqPlotController):
                    view.update_f(f)

    def on_update_nfft(self, nfft):
        for child in self.panel.rows:
            for pane in self.panel.cols[child]:
                view = self.panel.view[child][pane]
                if view is None:
                    continue
                if isinstance(view, FreqPlotController):
                    view.update_nfft(nfft)

    def _check_f(self):
        def _update_f():
            return (self.model.f[0], self.model.f[-1]+(self.model.f[-1]-self.model.f[-2]), len(self.model.f))
        if self._last_f is None:
            self._last_f = _update_f()
            self.on_update_f(self._last_f)
        elif not self.model.f[0] == self._last_f[0] and not len(self.model.f) == self._last_f[2]:
            self._last_f = _update_f()
            self.on_update_f(self._last_f)
        else:
            self.on_update_f(self._last_f)

    def draw(self):
        self.draw_tb()
        self.draw_ctrl()
        self.draw_view()

    def draw_tb(self):
        self.view.var_samp.set(self.model.reader.cur_samp)

        self.view.var_time_cur.set(strfmt_td(dt.timedelta(seconds=self.model.cur_time())))
        self.view.var_time_tot.set(strfmt_td(dt.timedelta(seconds=self.model.tot_time())))

        self.view.var_sweep.set(f"{self.model.sweep_time:02.3f}")
        self.view.var_show.set(f"{self.time_show:02.3f}")

    def draw_ctrl(self):
        self.view.var_file.set(str(self.model.reader.path))
        self.view.var_file_fmt.set(str(self.model.reader.fmt.name))

        self.view.var_fs.set(str(self.model.Fs))
        self.view.var_cf.set(str(self.model.cf))
        self.view.lbl_nfft.configure(text=str(self.model.nfft))
        self.view.var_nfft_exp.set(str(self.nfft_exp))

        self.view.lbl_block_size.configure(text=str(self.model.block_size))
        self.view.lbl_sweep_samples.configure(text=str(self.model.sweep_samples()))

    def draw_view(self):
        pass

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.view.ent_sweep:
            self.set_time_sweep(self.view.var_sweep.get())
        elif event.widget == self.view.ent_show:
            self.set_time_show(self.view.var_show.get())
        elif event.widget == self.view.cb_file_fmt:
            self.set_dtype(self.view.var_file_fmt.get())
        elif event.widget == self.view.ent_fs:
            self.set_fs(self.view.var_fs.get())
        elif event.widget == self.view.ent_cf:
            self.set_cf(self.view.var_cf.get())
        elif event.widget == self.view.ent_nfft_exp:
            self.set_nfft(self.view.var_nfft_exp.get())

    def handle_btn_file(self):
        self.set_path(dialog.get_file(False))

    def handle_sld_samp(self, *args):
        self.set_samp(self.view.var_samp.get())

    def set_samp(self, samp):
        self.stop()
        self.model.reader.cur_samp = samp
        self.draw_tb()
        # print(samp)
    def set_time_sweep(self, ts):
        try:
            self.model.set_sweep_time(float(ts))
        except ValueError:
            pass
        self.view.var_sweep.set(f"{self.model.sweep_time:02.3f}")
        self.draw_ctrl()
    def set_time_show(self, ts):
        try:
            ts = float(ts)
            self.time_show = ts
        except ValueError:
            pass
        self.view.var_show.set(f"{self.time_show:02.3f}")
    def set_path(self, path):
        self.model.reader.path = path
        print(f"Path set to '{self.model.reader.path}'")
        self.view.sld_samp.scale.config(from_=0, to=self.model.reader.max_samp) # resolution=self.model.block_size
        self.draw_tb()
        self.draw_ctrl()
    def set_dtype(self, dtype):
        self.model.reader.fmt = dtype
        self.draw_tb()
        self.draw_ctrl()
    def set_fs(self, fs):
        self.model.Fs = fs
        self.view.var_fs.set(str(self.model.Fs))
        self.draw_tb()
        self.draw_ctrl()
    def set_cf(self, cf):
        self.model.cf = cf
        self.view.var_cf.set(str(self.model.cf))
        self.draw_ctrl()

    def set_nfft(self, exp):
        exp = int(exp)
        self.nfft_exp = exp
        self.model.nfft = 2**exp
        self.view.var_nfft_exp.set(str(self.nfft_exp))
        self.view.lbl_nfft.config(text=str(self.model.nfft))
        self.on_update_nfft(self.model.nfft)
        self.draw_ctrl()



import importlib

def GetController(view, mode):
    return importlib.import_module(f".controller.{view.path}", "pyspecan").GetController(mode)

def ControllerArgs(view, mode):
    return importlib.import_module(f".controller.{view.path}", "pyspecan").ControllerArgs(mode)



import numpy as np

class PMF:
    def __init__(self, bins=100, scale=0.2):
        self.num_bin = bins
        self.num = 0
        self._count = 0
        self._scale = scale
        self.b_range = None

        self._val = np.zeros(bins)
        self._bin = np.zeros(bins+1)

    def update(self, samples):
        mag = np.abs(samples)
        if self.b_range is None:
            self.b_range = (
                np.min(mag)*(1 + self._scale),
                np.max(mag)*(1 + self._scale)
            )
        pmf, bins = np.histogram(mag, bins=self.num_bin, range=self.b_range)

        self._val += pmf
        self.num += 1
        self._count += len(samples)

        if not np.any(bins == self._bin):
            self._bin[:] = bins

    @property
    def mean(self):
        y = self.y
        mean = 0
        for i in range(0, self.num_bin):
            mean = mean + i * y[i]
        mean = mean / self._count
        return mean

    @property
    def var(self):
        y = self.y
        mean = self.mean
        var = 0
        for i in range(0, self.num_bin):
            var = var + y[i] * (i - mean)**2
        var = var / (self._count-1)
        return var

    @property
    def std(self):
        return np.sqrt(self.var)

    @property
    def x(self):
        return self._bin[:-1]

    @property
    def y(self):
        r_pmf = self._val.real / self.num
        i_pmf = self._val.imag / self.num

        return (r_pmf/np.max(r_pmf)) + 1j*(i_pmf/np.max(i_pmf))

    @property
    def x_lim(self):
        return (self.b_range[0], self.b_range[1]) # type: ignore

    @property
    def y_lim(self):
        return (0, 1)

class ComplexPMF(PMF):
    def __init__(self, bins=100, scale=0.2):
        super().__init__(bins, scale)
        self._val = self._val.astype(np.complex64)
        self._bin = self._bin.astype(np.complex64)

    def update(self, samples):
        if self.b_range is None:
            self.b_range = (
                np.min([np.min(samples.real), np.min(samples.imag)])*(1 + self._scale),
                np.max([np.max(samples.real), np.max(samples.imag)])*(1 + self._scale)
            )
        r_pmf, r_bins = np.histogram(samples.real, bins=self.num_bin, range=self.b_range, density=False)
        i_pmf, i_bins = np.histogram(samples.imag, bins=self.num_bin, range=self.b_range, density=False)

        self._val += r_pmf.astype(np.float32) + 1j*i_pmf.astype(np.float32)
        self.num += 1
        self._count += len(samples)

        bins = r_bins + 1j*i_bins
        if not np.any(bins == self._bin):
            self._bin[:] = bins



"""Controller for SWEPT mode"""
import argparse
import numpy as np

from .base import Controller
from .base import define_args as base_args
# from .arc.plot_base import define_args as freq_args
from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .swept import plots

class ModeConfig:
    psd = True
    spg = False

def args_swept(parser: argparse.ArgumentParser):
    ctrl = base_args(parser)
    # freq_args(parser)
    # mode = parser.add_argument_group("SWEPT mode")
    # mode.add_argument("--psd", action="store_false", help="show psd")
    # mode.add_argument("--spg", action="store_true", help="show spectrogram")

class PanelControllerSwept(PanelController):
    def set_settings(self, child, pane):
        pane.cb_view.config(values=list(k for k in plots.keys()))
        pane.cb_view.bind("<<ComboboxSelected>>", lambda e,c=child, p=pane: self.set_view(e,c,p))

    def set_view(self, e, child: PanelChild, pane: Panel):
        view = pane.var_view.get()
        if view in plots:
            if self.view[child][pane] is not None:
                pane.wgts = {}
                pane.sets = {}
                for ch in pane.fr_main.winfo_children():
                    ch.destroy()
                for ch in pane.settings.winfo_children():
                    ch.destroy()
            plot = plots[view](self.parent, pane)
            self.view[child][pane] = plot

class ControllerSwept(Controller):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.panel = PanelControllerSwept(self, self.view.panel)
        child = self.panel.rows[0]
        pane = self.panel.cols[child][0]
        pane.var_view.set("PSD")
        self.panel.set_view(None, child, pane)
        self.draw()



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import FreqPlotController

from ....utils import vbw as _vbw
from ....obj import Frequency
from ....utils.window import WindowLUT
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    ref_level = 0
    scale_div = 10.0
    vbw = 10.0
    window = "blackman"

def define_args(parser: argparse.ArgumentParser):
    parser.add_argument("-rl", "--ref_level", default=PlotConfig.ref_level, type=float, help="ref Level")
    parser.add_argument("-sd", "--scale_div", default=PlotConfig.scale_div, type=float, help="scale per division")
    parser.add_argument("-vb", "--vbw", default=PlotConfig.vbw, type=float, help="video bandwidth")
    parser.add_argument("-w", "--window", default=PlotConfig.window, choices=[k for k in WindowLUT.keys()], help="FFT window function")

class PSD(FreqPlotController):
    def __init__(self, parent, pane: Panel, **kwargs):
        self.psd_min = None
        self.psd_max = None
        self.fmin = None
        self.fmax = None
        if not "ref_level" in kwargs:
            kwargs["ref_level"] = PlotConfig.ref_level
        if not "scale_div" in kwargs:
            kwargs["scale_div"] = PlotConfig.scale_div
        if not "vbw" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        if not "window" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        super().__init__(parent, pane, **kwargs)
        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("psd", fig.add_subplot())

        self.plotter.ax("psd").ax.set_autoscale_on(False)
        self.plotter.ax("psd").ax.locator_params(axis="x", nbins=5)
        self.plotter.ax("psd").ax.locator_params(axis="y", nbins=10)
        self.plotter.ax("psd").ax.grid(True, alpha=0.2)

        self.set_y()
        self.update()

    def reset(self):
        self.psd_min = None
        self.psd_max = None

    def plot(self, freq, psd):
        self.plotter.ax("psd").ax.set_title("PSD")

        if self.pane.sets["show_max"].get() == 1:
            if self.psd_max is None:
                self.psd_max = np.repeat(-np.inf, len(psd))
            self.psd_max[psd > self.psd_max] = psd[psd > self.psd_max]
            line_max = self.plotter.ax("psd").plot(freq, self.psd_max, name="psd_max", color="r")
        else:
            line_max = None
        if self.pane.sets["show_min"].get() == 1:
            if self.psd_min is None:
                self.psd_min = np.repeat(np.inf, len(psd))
            self.psd_min[psd < self.psd_min] = psd[psd < self.psd_min]
            line_min = self.plotter.ax("psd").plot(freq, self.psd_min, name="psd_min", color="b")
        else:
            line_min = None
        line_psd = self.plotter.ax("psd").plot(freq, psd, name="psd", color="y")

        self._show_y_location(psd)
        self.update()

    def update_f(self, f):
        fmin, fmax, fnum = f
        if self.fmin == fmin and self.fmax == fmax:
            return
        else:
            self.fmin = fmin
            self.fmax = fmax
        psd_tick = np.linspace(fmin, fmax, 5)
        psd_text = [str(Frequency.get(f)) for f in psd_tick]
        self.plotter.ax("psd").set_xlim(fmin, fmax)
        self.plotter.ax("psd").ax.set_xticks(psd_tick, psd_text)
        self.update()

    def update_nfft(self, nfft):
        self.reset()

    def set_y(self):
        """Set plot ylimits"""
        self.plotter.ax("psd").set_ylim(self.y_btm, self.y_top)

    def draw_settings(self, row=0):
        row = super().draw_settings(row)
        var_psd_min = tk.IntVar(self.pane.settings, 1)
        chk_show_min = ttk.Checkbutton(self.pane.settings, onvalue=1, offvalue=0,variable=var_psd_min)
        chk_show_min.configure(command=self.toggle_psd_min)
        var_psd_max = tk.IntVar(self.pane.settings, 1)
        chk_show_max = ttk.Checkbutton(self.pane.settings, onvalue=1, offvalue=0, variable=var_psd_max)
        chk_show_max.configure(command=self.toggle_psd_max)

        self.pane.wgts["show_min"] = chk_show_min
        self.pane.sets["show_min"] = var_psd_min
        self.pane.wgts["show_max"] = chk_show_max
        self.pane.sets["show_max"] = var_psd_max

        ttk.Separator(self.pane.settings, orient=tk.HORIZONTAL).grid(row=row,column=0,columnspan=3, pady=5, sticky=tk.EW)
        row += 1
        ttk.Label(self.pane.settings, text="PSD").grid(row=row, column=0,columnspan=2)
        row += 1
        ttk.Label(self.pane.settings, text="Max Hold").grid(row=row, column=0)
        chk_show_max.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="Min Hold").grid(row=row, column=0)
        chk_show_min.grid(row=row, column=1)
        row += 1
        return row

    # --- GUI bind events and setters --- #
    def set_scale(self, scale):
        prev = float(self.scale)
        super().set_scale(scale)
        if not prev == self.scale:
            self.set_y()

    def set_ref_level(self, ref):
        prev = float(self.ref_level)
        super().set_ref_level(ref)
        if not prev == self.ref_level:
            self.set_y()

    def set_vbw(self, smooth):
        prev = float(self.vbw)
        super().set_vbw(smooth)
        if not prev == self.vbw:
            self.psd_min = None
            self.psd_max = None

    def toggle_psd_min(self):
        """Toggle PSD min-hold visibility"""
        art = self.plotter.ax("psd").art("psd_min")
        if art is None:
            return
        if self.pane.sets["show_min"].get() == 0:
            self.psd_max = None
            art.set_visible(False)
        else:
            art.set_visible(True)
        self.update()

    def toggle_psd_max(self):
        """Toggle PSD max-hold visibility"""
        art = self.plotter.ax("psd").art("psd_max")
        if art is None:
            return
        if self.pane.sets["show_max"].get() == 0:
            self.psd_min = None
            art.set_visible(False)
        else:
            art.set_visible(True)
        self.update()



"""Base Controllers for tkGUI Controller"""
import argparse

import tkinter as tk
from tkinter import ttk

import numpy as np

from .panels import Panel

from ...utils.window import WindowLUT

from ...backend.mpl.plot import _Plot, Plot, BlitPlot

def define_args(parser: argparse.ArgumentParser, cfg):
    parser.add_argument("-rl", "--ref_level", default=cfg.ref_level, type=float, help="ref Level")
    parser.add_argument("-sd", "--scale_div", default=cfg.scale_div, type=float, help="scale per division")
    parser.add_argument("-vb", "--vbw", default=cfg.vbw, type=float, help="video bandwidth")
    parser.add_argument("-w", "--window", default=cfg.window, choices=[k for k in WindowLUT.keys()], help="FFT window function")

class _PlotController:
    """Controller for view.tkGUI.GUIPlot"""
    __slots__ = (
        "parent", "pane", "plotter", "ready",
        "fr_canv"
    )
    def __init__(self, parent, pane: Panel, **kwargs):
        self.parent = parent
        self.pane = pane
        self.ready = False
        self.plotter: Plot = None # type: ignore

        self.fr_canv = ttk.Frame(pane.fr_main)
        self.fr_canv.pack(fill=tk.BOTH, expand=True)

    def update(self):
        """Update view plot"""
        self.plotter.update()

    def reset(self):
        """Reset plot view"""
        pass

    def plot(self, *args, **kwargs):
        """Update plot data"""
        raise NotImplementedError()

    def draw_settings(self, row=0):
        """Initialize settings panel"""
        raise NotImplementedError()

class FreqPlotController(_PlotController):
    """Controller for view.tkGUI.GUIFreqPlot"""
    __slots__ = (
        "window", "vbw", "scale", "ref_level",
        "lbl_lo", "lbl_hi"
    )
    def __init__(self, parent, pane: Panel, **kwargs):
        super().__init__(parent, pane)

        self.window =  kwargs.get("window", "blackman")
        self.vbw = kwargs.get("vbw", 10.0)
        self.scale = kwargs.get("scale_div", 10.0)
        self.ref_level = kwargs.get("ref_level", 0.0)
        self.draw_settings()

        self.set_ref_level(self.pane.sets["ref_level"].get())

        self.lbl_lo = ttk.Label(self.fr_canv, text="V")
        self.lbl_hi = ttk.Label(self.fr_canv, text="^")

    def update(self):
        self.plotter.canvas.draw()

    def update_f(self, f):
        """Set plot xticks and xlabels"""

    def update_nfft(self, nfft):
        """Update plot nfft"""

    def plot(self, freq, psd):
        raise NotImplementedError()

    @property
    def y_top(self):
        """Return plot maximum amplitude"""
        return self.ref_level
    @property
    def y_btm(self):
        """Return plot minimum amplitude"""
        return self.ref_level - (10*self.scale)

    def _show_y_location(self, psd):
        if np.all(psd < self.y_btm):
            self.lbl_lo.place(relx=0.2, rely=0.9, width=20, height=20)
        else:
            if self.lbl_lo.winfo_ismapped():
                self.lbl_lo.place_forget()
        if np.all(psd > self.y_top):
            self.lbl_hi.place(relx=0.2, rely=0.1, width=20, height=20)
        else:
            if self.lbl_hi.winfo_ismapped():
                self.lbl_hi.place_forget()

    def draw_settings(self, row=0):
        var_scale = tk.StringVar(self.pane.settings, str(self.scale))
        ent_scale = ttk.Entry(self.pane.settings, textvariable=var_scale, width=10)
        ent_scale.bind("<Return>", self.handle_event)

        var_ref_level = tk.StringVar(self.pane.settings, str(self.ref_level))
        ent_ref_level = ttk.Entry(self.pane.settings, textvariable=var_ref_level, width=10)
        ent_ref_level.bind("<Return>", self.handle_event)

        var_vbw = tk.StringVar(self.pane.settings, str(self.vbw))
        ent_vbw = ttk.Entry(self.pane.settings, textvariable=var_vbw, width=10)
        ent_vbw.bind("<Return>", self.handle_event)

        var_window = tk.StringVar(self.pane.settings, str(self.window))
        cb_window = ttk.Combobox(self.pane.settings, textvariable=var_window, width=9)
        cb_window.bind("<<ComboboxSelected>>", self.handle_event)

        self.pane.wgts["scale"] = ent_scale
        self.pane.sets["scale"] = var_scale
        self.pane.wgts["ref_level"] = ent_ref_level
        self.pane.sets["ref_level"] = var_ref_level
        self.pane.wgts["vbw"] = ent_vbw
        self.pane.sets["vbw"] = var_vbw
        self.pane.wgts["window"] = cb_window
        self.pane.sets["window"] = var_window

        ttk.Label(self.pane.settings, text="Scale/Div").grid(row=row, column=0)
        ent_scale.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="Ref Level").grid(row=row, column=0)
        ent_ref_level.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="VBW").grid(row=row, column=0)
        ent_vbw.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="Window").grid(row=row, column=0)
        cb_window.grid(row=row, column=1)
        row += 1
        return row

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.pane.wgts["scale"]:
            self.set_scale(self.pane.sets["scale"].get())
        elif event.widget == self.pane.wgts["ref_level"]:
            self.set_ref_level(self.pane.sets["ref_level"].get())
        elif event.widget == self.pane.wgts["vbw"]:
            self.set_vbw(self.pane.sets["vbw"].get())
        elif event.widget == self.pane.wgts["window"]:
            self.set_window(self.pane.sets["window"].get())

    def set_scale(self, scale):
        """set plot scale"""
        try:
            scale = float(scale)
            self.scale = scale
        except ValueError:
            scale = self.scale
        self.pane.sets["scale"].set(str(self.scale))

    def set_ref_level(self, ref):
        """Set plot ref level"""
        try:
            ref = float(ref)
            self.ref_level = ref
        except ValueError:
            ref = self.ref_level
        self.pane.sets["ref_level"].set(str(self.ref_level))

    def set_vbw(self, smooth):
        """Set plot vbw"""
        try:
            smooth = float(smooth)
            if smooth <= 0.0:
                smooth = 0.0
            self.vbw = smooth
        except ValueError:
            smooth = self.vbw
        self.pane.sets["vbw"].set(str(self.vbw))
    def set_window(self, window):
        """Set plot window function"""
        self.window = window



"""Controller for RT mode"""
import argparse
import time

import numpy as np

from ...utils import matrix
from ...backend.mpl.color import cmap

from .base import Controller
from .base import define_args as base_args
# from .arc.plot_base import define_args as freq_args
from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .rt import plots

class ModeConfig:
    x = 1001
    y = 600
    cmap = "hot"

def args_rt(parser: argparse.ArgumentParser):
    ctrl = base_args(parser)
    # freq_args(parser)
    mode = parser.add_argument_group("RT mode")
    mode.add_argument("--x", default=ModeConfig.x, type=int, help="histogram x pixels")
    mode.add_argument("--y", default=ModeConfig.y, type=int, help="histogram y pixels")
    mode.add_argument("--cmap", default=ModeConfig.cmap, choices=[k for k in cmap.keys()], help="histogram color map")

class PanelControllerRT(PanelController):
    def set_settings(self, child, pane):
        pane.cb_view.config(values=list(k for k in plots.keys()))
        pane.cb_view.bind("<<ComboboxSelected>>", lambda e,c=child, p=pane: self.set_view(e,c,p))

    def set_view(self, e, child: PanelChild, pane: Panel):
        view = pane.var_view.get()
        if view in plots:
            if self.view[child][pane] is not None:
                pane.wgts = {}
                pane.sets = {}
                for ch in pane.fr_main.winfo_children():
                    ch.destroy()
                for ch in pane.settings.winfo_children():
                    ch.destroy()
            plot = plots[view](self.parent, pane)
            self.view[child][pane] = plot

class ControllerRT(Controller):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.panel = PanelControllerRT(self, self.view.panel)
        child = self.panel.rows[0]
        pane = self.panel.cols[child][0]
        pane.var_view.set("PST")
        self.panel.set_view(None, child, pane)
        self.draw()

    def loop(self):
        while self.running:
            time_show = self.time_show/1000 # convert ms to s
            valid, ptime = self._next()
            if not valid or ptime is None:
                break
            wait = time_show-ptime
            if wait > 0:
                self.view.lbl_msg.configure(text="")
                time.sleep(wait)
            else:
                # self.model.skip_time(-wait)
                self.view.lbl_msg.configure(text="OVERFLOW")



"""File reader"""
import pathlib
from enum import Enum

import numpy as np
from numpy._typing._array_like import NDArray

from .. import err

def read_ci8(path, count, offset):
    samples = np.fromfile(path, offset=offset, count=count, dtype=np.int8)
    return samples.astype(np.float32).view(dtype=np.complex64)
def read_ci16(path, count, offset):
    samples = np.fromfile(path, offset=offset, count=count, dtype=np.int16)
    return samples.astype(np.float32).view(dtype=np.complex64)
def read_cf16(path, count, offset):
    samples = np.fromfile(path, offset=offset, count=count, dtype=np.float16)
    return samples.astype(np.float32).view(dtype=np.complex64)
def read_cf32(path, count, offset):
    samples = np.fromfile(path, offset=offset, count=count, dtype=np.float32)
    return samples.view(dtype=np.complex64)
def read_cf64(path, count, offset):
    samples = np.fromfile(path, offset=offset, count=count, dtype=np.float64)
    return samples.astype(np.float32).view(dtype=np.complex64)

class Format(Enum):
    ci8  = (2, read_ci8)
    ci16 = (4, read_ci16)
    cf16 = (4, read_cf16)
    cf32 = (8, read_cf32)
    cf64 = (16, read_cf64)

    def size(self):
        return self.value[0]

    def read(self, path, count: int, offset: int):
        return self.value[1](path, count, offset*self.size())

    @classmethod
    def choices(cls):
        return [inst.name for inst in cls]

class Reader:
    __slots__ = (
        "_path", "_fmt",
        "_fsize", "cur_samp", "max_samp"
    )
    def __init__(self, fmt, path):
        self._path: pathlib.Path = None # type: ignore
        self._fmt: Format = None # type: ignore
        self._fsize = 0
        self.cur_samp = 0
        self.max_samp = -1

        self.set_path(path)
        self.set_fmt(fmt)

    def show(self, ind=0):
        print(" "*ind + f"{self.percent():06.2f}% [{self.fmt.name}] {self.path}")
        print(" "*ind + f"{self.cur_samp}/{self.max_samp}")

    def get_path(self):
        return self._path
    def set_path(self, path):
        if path is None or path == "":
            return
        path = pathlib.Path(path)

        if not path.exists():
            return
        if not path == self._path:
            self._fsize = path.stat().st_size
            if not self._fmt is None:
                self.max_samp = self._fsize // self.fmt.size()
        self._path = path
        self.reset()
    path = property(get_path, set_path)

    def get_fmt(self):
        return self._fmt
    def set_fmt(self, fmt):
        self._fmt = Format[fmt]
        if not self._fsize == 0 and not self._fmt is None:
            self.max_samp = self._fsize // self.fmt.size()
    fmt = property(get_fmt, set_fmt)

    def reset(self):
        self.cur_samp = 0

    def next(self, count: int):
        count *= 2
        if self.cur_samp + count > self.max_samp:
            raise err.Overflow(f"{self.cur_samp}+{count} > {self.max_samp}")
        samps = self._read(count)
        self.cur_samp += count
        return samps

    def prev(self, count: int):
        count *= 2
        if self.cur_samp - count < 0:
            raise err.Overflow(f"{self.cur_samp}-{count} < 0")
        self.cur_samp -= count
        samps = self._read(count)
        return samps

    def _read(self, count: int):
        """Read <count> samples"""
        if self.path is None:
            return
        samps = self.fmt.read(self.path, count, self.cur_samp)
        return samps

    def forward(self, count: int):
        """Return <count> samples until EOF"""
        while self.cur_samp + count <= self.max_samp:
            # return self.next(count)
            yield self.next(count)

    def reverse(self, count: int):
        while self.cur_samp - count >= 0:
            # return self.prev(count)
            yield self.prev(count)

    def percent(self):
        """Return percent of file read"""
        return float(self.cur_samp/self.max_samp)*100

    def __call__(self, count):
        return self.forward(count)



import numpy as np

from .window import WindowLUT
from .vbw import vbw as _vbw

def psd(samples, nfft=1024, overlap=0.8, Fs=1.0, vbw=None, win="blackman"):
    win_func = WindowLUT[win]
    overlap = 1-overlap
    n_samp = len(samples)
    n_frames = int(n_samp/(nfft*overlap))
    out = np.zeros((nfft, n_frames))

    # print(f"stft for {n_samp} samples / {out.shape[1]}*{nfft} FFT @ overlap {overlap}")

    for i in range(n_frames):
        s_idx = int(i*(nfft*overlap))
        e_idx = s_idx + nfft
        # print(f"snip {i} = [{s_idx}:{e_idx}]")
        if e_idx > n_samp:
            segment = np.zeros(nfft, dtype=samples.dtype)
            f_idx = n_samp-s_idx
            segment[:f_idx] = samples[s_idx:]
            # segment[f_idx:] = samples[-f_idx:]
        else:
            segment = samples[s_idx:e_idx]

        _psd = np.fft.fft(segment*win_func(nfft))
        _psd = np.abs(_psd)
        _psd = _psd**2 / (nfft*Fs)
        _psd = 10.0*np.log10(_psd)
        _psd = np.fft.fftshift(_psd)
        if vbw is not None:
            _psd = _vbw(_psd, vbw)
        out[:,i] = _psd
    return out



import numpy as np

def _fill(matrix):
    matrix = np.ascontiguousarray(matrix, dtype=np.int8)
    rows, cols = matrix.shape
    mask = matrix != 0

    s_idx = np.argmax(mask, axis=0)
    # e_idx = rows - 1 - np.argmax(mask[::-1], axis=0)
    e_idx = rows - 1 - np.argmax(np.flip(mask, axis=0), axis=0)

    # empty = mask.sum(axis=0) == 0
    empty = ~mask.any(axis=0)
    s_idx[empty] = 0
    e_idx[empty] = 0

    diff = np.zeros_like(matrix, dtype=np.int8)
    col_idx = np.arange(cols, dtype=np.int32)

    # s_rows = s_idx + 1
    # s_valid = s_rows < e_idx

    # e_rows = e_idx
    # e_valid = e_idx > s_idx
    # diff[s_rows[s_valid], col_idx[s_valid]] += 1
    # diff[e_rows[e_valid], col_idx[e_valid]] -= 1

    valid = s_idx < e_idx
    s_rows = s_idx[valid] + 1
    e_rows = e_idx[valid]

    np.add.at(diff, (s_rows, col_idx[valid]), 1)
    np.add.at(diff, (e_rows, col_idx[valid]), -1)

    np.cumsum(diff, axis=0, out=diff)
    matrix += diff

    # print(out)
    # exit()
    # for i in range(matrix.shape[1]):
    #     wh = np.where(matrix[:,i]>0)[0]
    #     matrix[np.min(wh):np.max(wh),i] += 1
    # matrix[np.min(matrix[:,None]):np.max(matrix[:,None]),None] += 1
    return matrix

def dot(x: int, y: int, psds, yt: float, yb: float):
    """Dot Matrix
    Each PSD's x/y mapped to histogram x/y
    """
    hist = np.zeros((y, x), dtype=np.int8)

    if np.all(psds < yb):
        return hist
    elif np.all(psds > yt):
        return hist

    y_rng = abs(abs(yt) - abs(yb))
    y_off = yb if yb > 0 else -yb

    x_ratio = x/psds.shape[0]
    y_ratio = y/y_rng

    x_idx = np.floor(np.arange(0, psds.shape[0])*x_ratio).astype(int)
    y_idx = np.floor((psds+y_off)*y_ratio).astype(int)

    y_idx = np.clip(y_idx, 0, y-1)

    y_val = np.ones_like(y_idx)
    y_val[y_idx == 0] = 0
    y_val[y_idx >= y-1] = 0

    for i in range(psds.shape[1]):
        hist[y_idx[:,i], x_idx] += y_val[:,i]
    return hist

def cdot(x, y, psds, yt=0.05, yb=0.05):
    hist = dot(x, y, psds, yt, yb)
    return _fill(hist)

def vec(x, y, psds, yt=0.05, yb=0.05, interp=32):
    """Vector Matrix
    Interpolate between PSD x/y to mimic vector matrix
    """
    y_int = np.zeros((psds.shape[0]*interp, psds.shape[1]))
    t = np.arange(psds.shape[0])*interp
    t_int = np.arange(psds.shape[0]*interp)-1
    for i in range(psds.shape[1]):
        y_int[:,i] = np.interp(t_int, t, psds[:,i])
    return dot(x, y, y_int, yt, yb)

def cvec(x, y, psds, yt=0.05, yb=0.05, interp=16):
    hist = vec(x, y, psds, yt, yb, interp)
    return _fill(hist)



"""Configure ttk Style

TButton
TCheckbutton
TCombobox
TEntry
TFrame
TLabel
TLabelFrame
TMenubutton
TNotebook
TPanedwindow
Horizontal.TProgressbar
Vertical.TProgressbar
TRadiobutton
TScale
    Horizontal.TScale
    Vertical.TScale
TScrollbar
    Horizontal.TScrollbar
    Vertical.TScrollbar
TSeparator
TSizegrip
Treeview
"""
import tkinter as tk
from tkinter import ttk

def _print(style=None):
    if style is None:
        style = ttk.Style()
    print(style.theme_use(None))
    for element in ("TCheckbutton",):
        print(f"{element}:", end="")
        print(f" CONFIG {style.configure(element)} |", end="")
        print(f" MAP {style.map(element)} |", end="")
        print(f" LAYOUT {style.layout(element)} |", end="")
        print()
    exit()

class _Theme:
    """pyspecan theme"""
    _style: ttk.Style = None # type: ignore

    C_FG = "#FF00FF"
    C_FG_TEXT = C_FG
    C_FG_ALT = C_FG
    C_FG_ACTIVE = C_FG
    C_FG_DISABLED = C_FG
    C_FG_PRESSED = C_FG
    C_FG_SELECTED = C_FG
    C_FG_FOCUS = C_FG
    C_FG_READONLY = C_FG
    C_FG_INVALID = C_FG

    C_BG = "#00FF00"
    C_BG_TEXT = C_BG
    C_BG_ALT = C_BG
    C_BG_ACTIVE = C_BG
    C_BG_DISABLED = C_BG
    C_BG_PRESSED = C_BG
    C_BG_SELECTED = C_BG
    C_BG_FOCUS = C_BG
    C_BG_READONLY = C_BG
    C_BG_INVALID = C_BG

    def __new__(cls, parent):
        if cls._style is None:
            cls._style = ttk.Style(parent)
        # _print(cls._style)

        name = f"pyspecan.{cls.__name__}"

        try:
            cls._style.theme_create(name, "classic")
            cls._style.theme_use(name)
            cls._style.configure(".", font=("TKDefaultFont", 10))
            cls._config_base()
            cls._config()
        except tk.TclError: # The theme already exists
            cls._style.theme_use(name)
        # _print(cls._style)


    @classmethod
    def _config_base(cls):
        style = {
            "TButton": {
                "relief": "raised",
                "padding": (5,2)
            },
            "Settings.TButton": {
                "font": ("TkDefaultFont", 6),
                "padding": (5,0)
            },
            "Close.TButton": {
                "font": ("TkDefaultFont", 6),
                "padding": (5,0)
            },
            "AddRow.TButton": {
                "font": ("TkDefaultFont", 8),
                "padding": (5,0)
            },
            "AddCol.TButton": {
                "font": ("TkDefaultFont", 8),
                "padding": (5,0)
            },
        }
        cls.configure(style)
        maps = {
            "TButton": {
                "relief": [
                    ('disabled', 'flat')
                ]
            }
        }
        cls.map(maps)

    @classmethod
    def _config(cls):
        default_style = {
            "foreground": cls.C_FG,
            "selectforeground": cls.C_FG_TEXT,
            "fieldforeground": cls.C_FG_TEXT,
            "indicatorcolor": cls.C_FG,
            "background": cls.C_BG,
            "selectbackground": cls.C_BG_PRESSED,
            "fieldbackground": cls.C_BG_TEXT,
            "troughcolor": cls.C_BG_TEXT,
        }
        cls._style.configure(".", **default_style)
        default_maps = {
            "foreground": [
                ('active', cls.C_FG_ACTIVE),
                ('disabled', cls.C_FG_DISABLED),
            ],
            "background": [
                ('active', cls.C_BG_ACTIVE),
                ('disabled', cls.C_BG_DISABLED),
            ],
        }
        cls._style.map(".", **default_maps) # type: ignore
        style = {
            "TEntry": {
                "background": cls.C_BG_TEXT,
            },
            "TCheckbutton": {
                "indicatorcolor": cls.C_BG, # color when off
            },
        }
        cls.configure(style)
        maps = {
            "TButton": {
                "foreground":[
                    ('pressed', cls.C_FG_PRESSED),
                    ('disabled', cls.C_FG_DISABLED),
                ],
                "background": [
                    ('pressed', '!disabled', cls.C_BG_PRESSED),
                ]
            },
            "TEntry": {
                "fieldforeground": [
                    ('readonly', cls.C_FG_READONLY),
                    ('disabled', cls.C_FG_READONLY),
                ],
                "fieldbackground": [
                    ('readonly', cls.C_BG_READONLY),
                    ('disabled', cls.C_BG_READONLY),
                ],
            },
            "TCheckbutton": {
                "indicatorcolor": [
                    ("selected", cls.C_FG_PRESSED) # color when on
                ]
            }
        }
        cls.map(maps)

    @classmethod
    def configure(cls, obj, show=False):
        for k, v in obj.items():
            if not v:
                continue
            # if show:
            #     print(f"Configuring {k}, {v}")
            cls._style.configure(k, **v)
    @classmethod
    def map(cls, obj, show=False):
        for k, v in obj.items():
            if not v:
                continue
            # if show:
            #     print(f"Mapping {k}, {v}")
            cls._style.map(k, **v)



"""Error types"""
class _Error(Exception):
    """pyspecan Error"""

class UnknownOption(_Error):
    """Error thrown when an unknown option is provided"""

class Overflow(_Error):
    """Error thrown when an overflow occurs"""



import argparse

from .base import Model, define_args

from ..utils import stft

def args_rt(parser: argparse.ArgumentParser):
    define_args(parser)
    parser.add_argument("--overlap", default=0.6, type=float)
    parser.add_argument("--block_max", default=102400, type=int)

class ModelRT(Model):
    __slots__ = ("_overlap", "_block_max")
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._overlap = kwargs.get("overlap", 0.6)
        self._block_max = kwargs.get("block_max", 102400)
        self.update_blocksize()

    def psd(self, vbw=None, win="blackman"): # type: ignore
        if self._samples is None:
            return None
        if self._psd is None:
            if vbw is not None and vbw <= 0.0:
                vbw = None
            psd = stft.psd(self._samples, self.nfft, self.overlap, self.Fs.raw, vbw, win)
            self._psd = psd
        return self._psd

    def update_blocksize(self):
        self.block_size = int(self.Fs * (self._sweep_time/1000))
        if self.block_size > self._block_max:
            self.block_size = self._block_max
            super().set_sweep_time((self._block_size/self.Fs)*1000)

    def get_overlap(self):
        return self._overlap
    def set_overlap(self, overlap):
        if overlap <= 0.0 or overlap > 1.0:
            raise ValueError
        self._overlap = float(overlap)
    overlap = property(get_overlap, set_overlap)

    def get_sweep_time(self):
        return super().get_sweep_time()
    def set_sweep_time(self, ts):
        super().set_sweep_time(ts)
        self.update_blocksize()
    sweep_time = property(get_sweep_time, set_sweep_time)

    def get_fs(self):
        return super().get_fs()
    def set_fs(self, fs):
        super().set_fs(fs)
        self.update_blocksize()
    Fs = property(get_fs, set_fs)



"""Initialize pyspecan module/script"""
from . import err
from .config import config, Mode, View

from .model.model import GetModel, Model
from .view.view import GetView
from .controller.controller import GetController

class SpecAn:
    """Class to initialize pyspecan"""
    __slots__ = ("model", "view", "controller")
    def __init__(self, view, mode="psd", **kwargs):
        if config.PROFILE:
            from .utils.monitor import Profile
            Profile().enable()

        if config.MON_MEM:
            from .utils.monitor import Memory
            Memory().start()

        if not isinstance(mode, Mode):
            mode = Mode[mode]
            if mode == Mode.NONE:
                raise err.UnknownOption(f"Unknown mode {mode}")
        if not isinstance(view, View):
            view = View.get(view)
            if view == View.NONE:
                raise err.UnknownOption(f"Unknown view {view}")

        self.model: Model = GetModel(mode)(**kwargs)

        self.view = GetView(view, mode)(**kwargs)
        self.controller = GetController(view, mode)(self.model, self.view, **kwargs)

        self.model.show()
        self.view.mainloop()

        if config.MON_MEM:
            from .utils.monitor import Memory
            Memory().stop()

        if config.PROFILE:
            from .utils.monitor import Profile
            Profile().disable()
            if config.PROFILE_PATH is None:
                Profile().show()
            else:
                Profile().dump(config.PROFILE_PATH)



"""Configure matplotlib Style
"""
import matplotlib.pyplot as plt

theme = {
    "Light": lambda: plt.style.use("default"),
    "Dark": lambda: plt.style.use("dark_background"),
}

def get(name: str):
    return theme[name]



import datetime as dt

def split_td(td: dt.timedelta):
    s = td.total_seconds()
    hrs, remainder = divmod(s, 3600)
    mins, secs = divmod(remainder, 60)
    secs, ms = divmod(secs, 1)
    return int(hrs), int(mins), int(secs), int(ms*1000)

def strfmt_td(td: dt.timedelta):
    ch, cm, csec, cms = split_td(td)
    return f"{ch:03d}:{cm:02d}:{csec:02d}.{cms:03d}"



import numpy as np

def sigma2fwhm(sigma):
    return sigma * np.sqrt(8 * np.log(2))

def fwhm2sigma(fwhm):
    return fwhm / np.sqrt(8 * np.log(2))

def gaussian(y, sigma):
    fwhm = sigma2fwhm(sigma)
    x = np.arange(-int(sigma)*4, int(sigma)*4+1)

    kernel = np.exp(-(x) ** 2 / (2 * sigma ** 2))

    kernel_above_thresh = kernel > 0.0001
    finite_kernel = kernel[kernel_above_thresh]
    finite_kernel = finite_kernel / finite_kernel.sum()

    kernel_n_below_0 = int((len(finite_kernel) - 1) / 2.)

    smi = int(fwhm)
    smin = np.min(y)
    ex_samps = np.concatenate((np.repeat(smin, smi), y, np.repeat(smin, smi)))
    convolved_y = np.convolve(ex_samps, finite_kernel)
    smoothed_by_convolving = convolved_y[kernel_n_below_0+smi:(len(y)+kernel_n_below_0+smi)]
    return smoothed_by_convolving

def vbw(psd, smooth):
    return gaussian(psd, fwhm2sigma(smooth))



import numpy as np

from .window import WindowLUT
from .vbw import vbw as _vbw

def psd(samples, Fs=1.0, vbw=None, win="blackman"):
    N = len(samples)
    _psd = samples * WindowLUT[win](N)
    _psd = np.abs(np.fft.fft(_psd)) # type: ignore
    _psd = _psd**2 / (N*Fs)
    _psd = 10.0*np.log10(_psd)
    _psd = np.fft.fftshift(_psd)
    if vbw is not None:
        _psd = _vbw(_psd, vbw)
    return _psd



import tkinter as tk
from tkinter import ttk

from ...view.tkGUI.panels import PanelView, PanelChild, Panel

class PanelController:
    def __init__(self, parent, panel: PanelView):
        self.parent = parent
        self.panel = panel
        self.rows = []
        self.cols = {}
        self.view = {}

        self.panel.btn_row.configure(command=self.add_row)
        self.add_row()

    def add_row(self):
        frame = ttk.LabelFrame(self.panel.main, text=f"Row {len(self.rows)}")
        child = PanelChild(self, frame)
        self.rows.append(child)
        self.cols[child] = []
        self.view[child] = {}

        self.panel.main.add(frame, weight=1)

        child.btn_close.configure(command=lambda c=child: self.del_row(c))
        child.btn_col.configure(command=lambda c=child: self.add_col(c))

        self.add_col(child)
        self.panel.update_layout()

    def del_row(self, child: PanelChild):
        idx = self.rows.index(child)
        del self.cols[child]
        self.panel.main.remove(child.master)
        self.rows.pop(idx).root.destroy()

    def add_col(self, child: PanelChild):
        frame = ttk.LabelFrame(child.main, text=f"Col {len(self.cols[child])}")
        pane = Panel(self, frame)
        self.cols[child].append(pane)
        self.view[child][pane] = None

        child.main.add(frame, weight=1)
        #ttk.Label(self.panes[-1].main, text=f"Row {len(self.parent.panes)}, Col {len(self.panes)}").pack()

        pane.btn_close.configure(command=lambda c=child, p=pane: self.del_col(c,p))
        pane.btn_toggle.configure(command=lambda c=child, p=pane: self.toggle_settings(c,p))
        self.toggle_settings(child, pane)

        self.set_settings(child, pane)
        child.update_layout()

    def del_col(self, child: PanelChild, pane: Panel):
        idx = self.cols[child].index(pane)
        child.main.remove(pane.master)
        self.cols[child].pop(idx).root.destroy()
        if len(self.cols[child]) == 0:
            self.del_row(child)

    def toggle_settings(self, child: PanelChild, pane: Panel):
        """Toggle settings panel visibility"""
        if pane.fr_sets.winfo_ismapped():
            pane.fr_sets.forget()
            # self.btn_toggle.config(text="Show Settings")
        else:
            pane.fr_sets.pack(side=tk.LEFT, fill=tk.Y, before=pane.fr_main)
            # self.btn_toggle.config(text="Hide Settings")

    def set_settings(self, child: PanelChild, pane: Panel):
        raise NotImplementedError

    def set_view(self, e, child: PanelChild, pane: Panel):
        raise NotImplementedError

    def reset(self):
        pass

    def get_pane(self, child: PanelChild, pane: Panel):
        return self.cols[child][self.cols[child].index(pane)]

    def get_view(self, child: PanelChild, pane: Panel):
        return self.view[child][pane]



import typing
import numpy as np

from .. import err
from ..config import config, Mode
from ..obj import Frequency
from ..utils.window import WindowLUT

from .reader import Reader, Format

def define_args(parser):
    parser.add_argument("-f", "--path", default=None, help="file path")
    parser.add_argument("-d", "--fmt", choices=Format.choices(), default=Format.cf32.name, help="data format")

    parser.add_argument("-fs", "--Fs", default=1, type=Frequency.get, help="sample rate")
    parser.add_argument("-cf", "--cf", default=0, type=Frequency.get, help="center frequency")
    parser.add_argument("-n", "--nfft", default=1024, help="FFT size")
    parser.add_argument("-st", "--sweep_time", default=50.0, help="[ms] fft sweep time")

class Model:
    __slots__ = (
        "mode", "reader",
        "f", "_samples", "_psd", "_forward", "_reverse",
        "_Fs", "_cf", "_nfft",
        "_block_size", "_sweep_time"
    )
    def __init__(self, **kwargs):
        path = kwargs.get("path", None)
        fmt = kwargs.get("fmt", None)
        Fs = kwargs.get("Fs", 1)
        cf = kwargs.get("cf", 1)
        nfft = kwargs.get("nfft", 1024)
        sweep_time = kwargs.get("sweep_time", 50.0)

        self.reader = Reader(fmt, path)
        self._Fs = Frequency.get(Fs)
        self._cf = Frequency.get(cf)

        self._nfft = int(nfft)

        self.f = np.arange(-self._Fs.raw/2, self._Fs.raw/2, self._Fs.raw/self._nfft) + self._cf.raw
        self._samples = np.empty(self._nfft, dtype=np.complex64)
        self._psd = np.empty(self._nfft, dtype=np.float32)
        self._block_size = self._nfft
        self._sweep_time = float(sweep_time)

    def show(self, ind=0):
        print(" "*ind + f"{type(self).__name__} Reader:")
        self.reader.show(ind+2)

    def reset(self):
        self.reader.reset()

    @property
    def samples(self):
        return self._samples

    def psd(self, vbw=None, win="blackman") -> np.ndarray:
        ...

    def next(self):
        try:
            samples = self.reader.next(self._block_size)
        except err.Overflow:
            return False
        self._samples = samples
        self._psd = None
        return True

    def prev(self):
        try:
            samples = self.reader.prev(self._block_size)
        except err.Overflow:
            return False
        self._samples = samples
        self._psd = None
        return True

    def cur_time(self):
        return self.reader.cur_samp/self.Fs

    def tot_time(self):
        return self.reader.max_samp/self.Fs

    def skip_time(self, s):
        samps = int(self.Fs * s)
        # print(f"Skipping {s:.3f}s, {samps} ({samps/self.reader.max_samp*100:.2f}%)")
        self.reader.cur_samp += samps

    def get_fs(self):
        return self._Fs
    def set_fs(self, fs):
        self._Fs = Frequency.get(fs)
    Fs = property(get_fs, set_fs)

    def get_cf(self):
        return self._cf
    def set_cf(self, cf):
        self._cf = Frequency.get(cf)
    cf = property(get_cf, set_cf)

    def get_nfft(self):
        return self._nfft
    def set_nfft(self, nfft):
        self._nfft = int(nfft)
        self.f = np.arange(-self._Fs.raw/2, self._Fs.raw/2, self._Fs.raw/self._nfft) + self._cf.raw
        self._psd = np.empty(self._nfft, dtype=np.float32)
    nfft = property(get_nfft, set_nfft)

    def get_block_size(self):
        return self._block_size
    def set_block_size(self, size):
        self._block_size = size
        self._samples = np.empty(self._block_size, dtype=np.complex64)
    block_size = property(get_block_size, set_block_size)

    def get_sweep_time(self):
        return self._sweep_time
    def set_sweep_time(self, ts):
        self._sweep_time = ts
    sweep_time = property(get_sweep_time, set_sweep_time)

    def sweep_samples(self):
        return int(self.Fs * (self.sweep_time/1000))



from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.pyplot as plt

from .ax import Ax, BlitAx

# plt.style.use('dark_background')

class _Plot:
    """Generic plotting parent"""
    __slots__ = ("_fig", "_canv", "_axs")
    def __init__(self, fig: Figure = plt.figure()):
        self._fig = fig
        self._canv = fig.canvas
        self._axs = {}

    def show(self):
        print(f"axs: {self._axs}")

    def update(self):
        """Update plot"""
        self._canv.draw()

    @property
    def fig(self):
        return self._fig

    @property
    def canvas(self):
        return self._canv

    def ax(self, name) -> Ax:
        return self._axs[name]

    def axes(self):
        return [v for v in self._axs.values()]

    def add_ax(self, name, ax: Axes):
        """Add new axes"""
        self._axs[name] = Ax(ax)

    def cla(self, name=None):
        if name is None:
            for xname, ax in self._axs.items():
                ax.cla()
        else:
            self.ax(name).cla()

class Plot(_Plot):
    def __init__(self, fig: Figure = plt.figure()):
        super().__init__(fig)
        for i, ax in enumerate(fig.axes):
            self._axs[i] = Ax(ax, self)

class BlitPlot(_Plot):
    """Plot supporting blitting"""
    __slots__ = ("_bg", "_cid")
    def __init__(self, fig: Figure = plt.figure()):
        super().__init__(fig)
        for i, ax in enumerate(fig.axes):
            self._axs[i] = BlitAx(ax, self)
        self._bg = None
        self._cid = self._canv.mpl_connect("draw_event", self._on_draw)

    def update(self):
        cv = self.canvas
        fig = cv.figure
        if self._bg is None:
            self._on_draw(None)
        else:
            cv.restore_region(self._bg) # type: ignore
            self._draw_animated()
            # cv.blit(self._fig.bbox)
            cv.blit(fig.bbox)
        cv.flush_events()

    def add_ax(self, name, ax: Axes):
        self._axs[name] = BlitAx(ax)

    def _on_draw(self, event):
        cv = self.canvas
        if event is not None:
            if event.canvas != cv:
                raise RuntimeError
        self._bg = cv.copy_from_bbox(cv.figure.bbox) # type: ignore
        self._draw_animated()

    def _draw_animated(self):
        for xname, ax in self._axs.items():
            for aname, art in ax.artists.items():
                # print(f"Drawing {xname}[{aname}]")
                self._fig.draw_artist(art)



from typing import Callable

from .. import err
from ..config import config, Mode

from .base import Model
from .modes import ModelSwept, ModelRT
from .modes import args_swept, args_rt

def GetModel(mode):
    if mode == Mode.SWEPT:
        return ModelSwept
    elif mode == Mode.RT:
        return ModelRT
    else:
        raise err.UnknownOption(f"Unknown mode specified: {mode}")

def ModelArgs(mode) -> Callable:
    if mode == Mode.SWEPT:
        return args_swept
    elif mode == Mode.RT:
        return args_rt
    else:
        raise err.UnknownOption(f"Unknown mode specified: {mode}")



import argparse

from .base import Model, define_args

from ..utils import psd as _psd

def args_swept(parser: argparse.ArgumentParser):
    define_args(parser)

class ModelSwept(Model):
    def psd(self, vbw=None, win="blackman"): # type: ignore
        if self._samples is None:
            return None
        if self._psd is None:
            if vbw is not None and vbw <= 0.0:
                vbw = None
            psd = _psd.psd(self._samples, self.Fs.raw, vbw, win)
            self._psd = psd
        return self._psd

    def next(self):
        if self.sweep_time <= 0.0:
            return super().next()
        if super().next():
            self.reader.cur_samp += int(self.Fs * (self.sweep_time/1000))
            return True
        return False

    def prev(self):
        if self.sweep_time <= 0.0:
            return super().prev()
        if super().prev():
            self.reader.cur_samp -= int(self.Fs * (self.sweep_time/1000))
            return True
        return False

    def get_nfft(self):
        return super().get_nfft()
    def set_nfft(self, nfft):
        super().set_nfft(nfft)
        self.block_size = self._nfft
    nfft = property(get_nfft, set_nfft)



"""pyspecan

This file enables using `python3 -m pyspecan`

This makes specan available, by using _internal.main
Use `python3 -m pyspecan --help` to see available arguments
"""
import sys
import pathlib

if __name__ == "__main__":
    try:
        from pyspecan._internal.main import main as _main
    except ImportError: # Makes import work when running src/pyspecan without it installed
        sys.path.insert(0, str(pathlib.Path(__file__).parent.parent))
        from pyspecan._internal.main import main as _main
    sys.exit(_main())



"""Create a GUI view"""
import tkinter as tk
import tkinter.ttk as ttk

from ..base import View as _View
from ...config import config, Mode

from ...backend.tk import widgets
from ...backend.mpl import theme as theme_mpl

from .panels import PanelView

class View(_View):
    """Parent tkGUI view class"""
    def __init__(self, root=tk.Tk(), **kwargs):
        self.root = root
        if config.MODE == Mode.SWEPT:
            self.root.title("pyspecan | Swept")
        elif config.MODE == Mode.RT:
            self.root.title("pyspecan | Real-Time")

        theme_mpl.get(kwargs.get("theme", "Dark"))() # Set matplotlib theme

        # self.style = ttk.Style(root)
        self.root.title(f"pyspecan")
        self.root.protocol("WM_DELETE_WINDOW", self.quit)

        self._main = ttk.Frame(self.root)
        self._main.pack(expand=True, fill=tk.BOTH)

        self.fr_tb = ttk.Frame(self._main, height=20)
        self.draw_tb(self.fr_tb)
        self.fr_tb.pack(side=tk.TOP, fill=tk.X)

        self.main = ttk.PanedWindow(self._main, orient=tk.HORIZONTAL)
        self.main.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        self.fr_view = ttk.Frame(self.main)
        self.panel = PanelView(self.fr_view)

        self.fr_ctrl = ttk.Frame(self.main, width=100)
        self.draw_ctrl(self.fr_ctrl)
        self.main.add(self.fr_ctrl)

        self.main.add(self.fr_view)

    def draw_tb(self, parent):
        """Draw toolbar frame"""
        col = 0
        self.var_samp = tk.IntVar(parent)
        self.sld_samp = widgets.Scale(
            parent, variable=self.var_samp, length=150
        )
        self.sld_samp.grid(row=0,rowspan=2,column=col, sticky=tk.NSEW)
        col += 1
        self.var_time_cur = tk.StringVar(parent)
        self.var_time_tot = tk.StringVar(parent)
        self.lbl_time_cur = ttk.Label(parent, textvariable=self.var_time_cur)
        self.lbl_time_cur.grid(row=0,column=col)
        self.lbl_time_tot = ttk.Label(parent, textvariable=self.var_time_tot)
        self.lbl_time_tot.grid(row=1,column=col)
        col += 1
        ttk.Separator(parent, orient=tk.VERTICAL).grid(row=0,rowspan=2,column=col, padx=5, sticky=tk.NS)

        col += 1
        ttk.Label(parent, text="Sweep").grid(row=0,column=col)
        self.var_sweep = tk.StringVar(parent)
        self.ent_sweep = ttk.Entry(parent, textvariable=self.var_sweep, width=8)
        self.ent_sweep.grid(row=1,column=col, padx=2, pady=2)
        col += 1
        ttk.Label(parent, text="Show").grid(row=0,column=col)
        self.var_show = tk.StringVar(parent)
        self.ent_show = ttk.Entry(parent, textvariable=self.var_show, width=8)
        self.ent_show.grid(row=1,column=col, padx=2, pady=2)

        col += 1
        ttk.Separator(parent, orient=tk.VERTICAL).grid(row=0,rowspan=2,column=col, padx=5, sticky=tk.NS)
        col += 1
        self.btn_prev = ttk.Button(parent, text="Prev")
        self.btn_prev.grid(row=0,rowspan=2,column=col, padx=2,pady=2)
        col += 1
        self.btn_next = ttk.Button(parent, text="Next")
        self.btn_next.grid(row=0,rowspan=2,column=col, padx=2,pady=2)
        col += 1
        self.btn_start = ttk.Button(parent, text="Start")
        self.btn_start.grid(row=0,rowspan=2,column=col, padx=2,pady=2, sticky=tk.NS)
        col += 1
        self.btn_stop = ttk.Button(parent, text="Stop", state=tk.DISABLED)
        self.btn_stop.grid(row=0,rowspan=2,column=col, padx=2,pady=2, sticky=tk.NS)
        col += 1
        self.btn_reset = ttk.Button(parent, text="Reset")
        self.btn_reset.grid(row=0,rowspan=2,column=col, padx=2,pady=2, sticky=tk.NS)
        col += 1
        ttk.Separator(parent, orient=tk.VERTICAL).grid(row=0,rowspan=2,column=col, padx=5, sticky=tk.NS)

        col += 1
        self.lbl_msg = ttk.Label(parent, text="")
        self.lbl_msg.grid(row=0,column=col, rowspan=2, sticky=tk.E)
        parent.grid_columnconfigure(col, weight=1)
        col += 1
        self.var_draw_time = tk.StringVar(parent)
        self.lbl_draw_time = ttk.Label(parent, textvariable=self.var_draw_time)
        ttk.Label(parent, text="Draw").grid(row=0,column=col, sticky=tk.E)
        self.lbl_draw_time.grid(row=1,column=col, sticky=tk.E)
        parent.grid_columnconfigure(col, weight=1)

    def draw_ctrl(self, parent):
        """Draw control frame"""
        # Model
        root = ttk.Frame(parent) # File reader
        root.columnconfigure(2, weight=1)
        row = 0
        self.var_file = tk.StringVar(root)
        self.btn_file = ttk.Button(root, text="File")
        self.btn_file.grid(row=row,column=0, sticky=tk.W)
        self.ent_file = ttk.Entry(root, textvariable=self.var_file, state=tk.DISABLED, width=10)
        self.ent_file.grid(row=row,column=1,columnspan=2, sticky=tk.NSEW)
        row += 1
        ttk.Label(root, text="Format:").grid(row=row,column=0,sticky=tk.W)
        self.var_file_fmt = tk.StringVar(root)
        self.cb_file_fmt = ttk.Combobox(root, textvariable=self.var_file_fmt, width=5)
        self.cb_file_fmt.grid(row=row,column=1, sticky=tk.W)
        root.pack(padx=2,pady=2, fill=tk.X)

        root = ttk.Frame(parent) # File params
        row = 0
        self.var_fs = tk.StringVar(root)
        ttk.Label(root, text="Sample rate:").grid(row=row,column=0, sticky=tk.W)
        self.ent_fs = ttk.Entry(root, textvariable=self.var_fs, width=10)
        self.ent_fs.grid(row=row,column=1, sticky=tk.W)
        row += 1
        self.var_cf = tk.StringVar(root)
        ttk.Label(root, text="Center freq:").grid(row=row,column=0, sticky=tk.W)
        self.ent_cf = ttk.Entry(root, textvariable=self.var_cf, width=10)
        self.ent_cf.grid(row=row,column=1, sticky=tk.W)
        root.pack(padx=2,pady=2, fill=tk.X)

        root = ttk.Frame(parent) # Other params
        root.columnconfigure(2, weight=1)
        row = 0
        self.var_nfft_exp = tk.StringVar(root)
        ttk.Label(root, text="NFFT 2^").grid(row=row,column=0, sticky=tk.W)
        self.ent_nfft_exp = ttk.Entry(root, textvariable=self.var_nfft_exp, width=2)
        self.ent_nfft_exp.grid(row=row,column=1, sticky=tk.W)
        self.lbl_nfft = ttk.Label(root)
        self.lbl_nfft.grid(row=row,column=3, sticky=tk.NSEW)
        root.pack(padx=2,pady=2, fill=tk.X)

        ttk.Separator(parent, orient=tk.HORIZONTAL).pack(fill=tk.X, pady=5)

        root = ttk.Frame(parent) # Useful variables
        row = 0
        ttk.Label(root, text="Block size: ").grid(row=row, column=0, sticky=tk.W)
        self.lbl_block_size = ttk.Label(root)
        self.lbl_block_size.grid(row=row, column=1)
        row += 1
        ttk.Label(root, text="Sweep size: ").grid(row=row, column=0, sticky=tk.W)
        self.lbl_sweep_samples = ttk.Label(root)
        self.lbl_sweep_samples.grid(row=row, column=1)
        root.pack(padx=2,pady=2, fill=tk.X)


    def mainloop(self):
        self.root.mainloop()

    def quit(self):
        self.root.quit()
        self.root.destroy()
